//-< TESTDB.CXX >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     11-Aug-2000  K.A. Knizhnik  * / [] \ *
//                          Last update: 11-Aug-2000  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Example of using wide character strings and date types
//-------------------------------------------------------------------*--------*

#include "goods.h"
#include "datetime.h"

#define INPUT_BUF_SIZE 256 

wchar_t* input(char* prompt, wchar_t* buf, size_t buf_size)
{
    wchar_t* p;
    do { 
	printf(prompt);
	*buf = '\0';
	fgetws(buf, buf_size, stdin);
	p = buf + wcslen(buf);
    } while (p <= buf+1); 
    
    if (*(p-1) == '\n') {
	*--p = '\0';
    }
    return buf;
}

int inputInt(char* prompt) { 
    wchar_t buf[64];
    int val;
    while (swscanf(input(prompt, buf, sizeof buf), L"%d", &val) != 1);
    return val;
}


dbDateTime inputDateTime(char* prompt) { 
    wchar_t buf[64];
    int year, month, day;
    while (swscanf(input(prompt, buf, sizeof buf), L"%d.%d.%d", &year, &month, &day) != 3);
    return dbDateTime(year, month, day);
}

dbDate inputDate(char* prompt) { 
    wchar_t buf[64];
    int year, month, day;
    while (swscanf(input(prompt, buf, sizeof buf), L"%d.%d.%d", &year, &month, &day) != 3);
    return dbDate(year, month, day);
}



class Company : public object { 
  public:
    ref<Company> next;
    wstring_t    name;
    wstring_t    address;
    dbDate       founded;
    int4         income;
    wstring_t    CEO;
    int4         finance;    
    wstring_t    extraInfo[5];
    dbDateTime   inspectionDate;

    Company() : object(self_class) {}

    void dump() const { 
 	char buf[2][INPUT_BUF_SIZE];
	printf("Name: %S\nAddress: %S\nFounded: %s\nIncome: %d\nCEO: %S\nFinance: %d\n"
	       "Info[1]: %S\nInfo[2]: %S\nInspection date: %s\n\n", 
	       name.getWideChars(), address.getWideChars(), founded.asString(buf[0]),
	       income, CEO.getWideChars(), finance, (wchar_t*)extraInfo[0], (wchar_t*)extraInfo[1],
	       inspectionDate.asString(buf[1], sizeof(buf[1])));
    }
	       
    METACLASS_DECLARATIONS(Company, object);
};


class Corporation : public object { 
  public:
    wstring_t    name;
    ref<Company> companies;
    
    void addCompany(Company* company) { 
	company->next = companies;
	companies = company;
    }
    
    void initialize() const { 
	if (is_abstract_root()) { 
	    wchar_t buf[INPUT_BUF_SIZE];
	    ref<Corporation> root = this;
	    modify(root)->become(new Corporation(input("Corporation name:", buf, sizeof buf)));
	} 
    }
    
    Corporation(wchar_t* str) : object(self_class) { 
	name = str;
    }
    
    METACLASS_DECLARATIONS(Corporation, object);
};

static event updateEvent, terminateEvent;
static bool  shutdownFlag;

void task_proc listener(void*)
{
    while (!shutdownFlag) {
	updateEvent.wait();
	updateEvent.reset();
	if (!shutdownFlag) {
	    printf("Database updated\n");
	}
    }
    terminateEvent.signal();
}

int main() 
{
    database db;
    task::initialize(task::huge_stack);
    if (db.open("testdb.cfg")) { 
	ref<Corporation> root;
	db.get_root(root);
	root->initialize();
	root->signal_on_modification(updateEvent);
	task::create(listener, NULL);

	while (true) { 
	    wchar_t buf[INPUT_BUF_SIZE];
	    printf("Menu:\n1. Add company\n2. List companies.\n3. Exit\n\n");
	    input(">> ", buf, sizeof buf);
	    switch (*buf) { 
	      case '1':
		{ 
		    Company* company = new Company();
		    company->name = input("Name: ", buf, sizeof buf);
		    company->address = input("Address: ", buf, sizeof buf);
		    company->founded = inputDate("Founded [YYYY.MM.DD]: ");
		    company->income = inputInt("Income: ");
		    company->CEO = input("CEO: ", buf, sizeof buf);
		    company->finance = inputInt("Finance: ");
		    company->extraInfo[0] = input("Extra information 1: ", buf, sizeof buf);
		    company->extraInfo[1] = input("Extra information 2: ", buf, sizeof buf);
		    company->inspectionDate = inputDateTime("Inspection date [YYYY.MM.DD]: ");
		    modify(root)->addCompany(company);
		}
		break;
	      case '2':
		{ 
		    for (ref<Company> company = root->companies; !company.is_nil(); company = company->next) 
		    {
			company->dump();
		    }
		}
		printf("-------------------\n");
		break;
	      case '3':
		printf("Session closed\n");
		shutdownFlag = true;
		updateEvent.signal();
		terminateEvent.wait();
		db.close();
		return 0;
	    }
	}
	return 0;
    } else { 
	fprintf(stderr, "Failed to open database\n");
	return 1;
    }
}

field_descriptor& Company::describe_components()
{
    return FIELD(next), FIELD(name), FIELD(address), FIELD(founded), FIELD(income), 
	FIELD(CEO), FIELD(finance), ARRAY(extraInfo), FIELD(inspectionDate); 
}

field_descriptor& Corporation::describe_components()
{
    return FIELD(name), FIELD(companies);
}

    
REGISTER(Company, object, pessimistic_repeatable_read_scheme);
REGISTER(Corporation, object, pessimistic_repeatable_read_scheme);
