//-< TestCol.java >--------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:     6-Oct-2000   K.A. Knizhnik  * / [] \ *
//                          Last update: 6-Oct-2000   K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test program for persistent collection classes
//-------------------------------------------------------------------*--------*

import goodsjpi.*;
import goodslib.*;
import java.util.*;

class DbRecord extends Persistent implements Comparable { 
    String value;
 
    public int compareTo(Object o) { 
	if (o instanceof DbRecord) { 
	    return value.compareTo(((DbRecord)o).value);
	}
	throw new ClassCastException();
    }

    public boolean equals(Object o) { 
	if (o instanceof DbRecord) { 
	    return value.equals(((DbRecord)o).value);
	}
	return false;
    }
    
    public int hashCode() { 
	return value.hashCode();
    }

    DbRecord(String v) { 
	value = v;
    }
}

public class TestCol extends Persistent { 
    PersistentArrayList pal;
    PersistentHashMap   phm;
    PersistentTreeMap   ptm;
    PersistentHashSet   phs;
    PersistentTreeSet   pts;

    final static int nNodes = 10000;
    final static int nArrayElems = 1000;
    
    void insert(String key) { 
	DbRecord rec = new DbRecord("Rec" + key);
	phm.put("Key" + key, rec);
	ptm.put("Key" + key, rec);
	phs.add(rec);
	pts.add(rec);
    }

    TestCol() { 
	pal = new PersistentArrayList();
	phm = new PersistentHashMap();
	ptm = new PersistentTreeMap();
	phs = new PersistentHashSet();
	pts = new PersistentTreeSet();
    }

    public static void main(String args[]) throws Exception { 
	Database db = new Database();
	if (db.open("testcol.cfg")) { 	    
	    long start = System.currentTimeMillis();
	    TestCol root = (TestCol)db.getRoot();
	    if (root == null) { 
		root = new TestCol();
		db.setRoot(root);
	    }
	    runTest(root);
	    //root.runTestTransaction();
	    db.close();
	    System.out.println("Elapsed time " + (System.currentTimeMillis() - start) + " msec");
	}
    }

    static void runTest(TestCol root) { 
	int i;
	DbRecord rec;
	Map.Entry entry;

	String key[] = new String[nNodes];
	Random rnd = new Random(2001);
	for (i = 0; i < nNodes; i++) { 
	    key[i] = Long.toString(rnd.nextLong());
	}
	for (i = 0; i < nArrayElems; i++) { 
	    root.pal.add(new DbRecord(Integer.toString(i)));
	}
	for (i = 0; i < nNodes; i++) { 
	    root.insert(key[i]);
	    if (i % 1000 == 0) { 
		System.out.print("\rInsert " + i + " nodes");
	    }
	}
	System.out.println("\nDatabase initialized");
			   	
	Assert.that(root.pal.size() == nArrayElems);
	Assert.that(root.phm.size() == nNodes);
	Assert.that(root.ptm.size() == nNodes);
	Assert.that(root.phs.size() == nNodes);
	Assert.that(root.pts.size() == nNodes);

	for (i = 0; i < nArrayElems; i++) { 
	    Assert.that(((DbRecord)root.pal.get(i)).value.equals(Integer.toString(i)));
	}
	for (i = 0; i < nNodes; i++) { 
	    rec = (DbRecord)root.ptm.get("Key" + key[i]);
	    Assert.that(root.phm.get("Key" + key[i]) == rec);
	    Assert.that(root.pts.contains(rec));
	    Assert.that(root.phs.contains(rec));
	}	
	System.out.println("Random search test passed");
	
	Arrays.sort(key);

	Iterator i1, i2, i3, i4;
	i1 = root.phm.entrySet().iterator();
	i2 = root.phm.keySet().iterator();
	i3 = root.phm.values().iterator();
	i4 = root.phs.iterator();
	boolean mark[] = new boolean[key.length];
	
	for (i = 0; i < nNodes; i++) { 
	    entry = (Map.Entry)i1.next();
	    Assert.that(i2.next() == entry.getKey());
	    Assert.that(i3.next() == entry.getValue());
	    String k = entry.getKey().toString().substring(3);
	    int j = Arrays.binarySearch(key, k);
	    Assert.that(j >= 0);
	    Assert.that(!mark[j]);
	    Assert.that(entry.getKey().equals("Key" + key[j]));
	    Assert.that(((DbRecord)entry.getValue()).value.equals("Rec" + key[j]));
	    Assert.that(i4.next() != null);
	    mark[j] = true;
	    i4.remove();
	}
	Assert.that(!i1.hasNext());
	Assert.that(!i2.hasNext());
	Assert.that(!i3.hasNext());
	Assert.that(!i4.hasNext());
	System.out.println("Hash iterators test passed");
	i1 = root.ptm.entrySet().iterator();
	i2 = root.ptm.keySet().iterator();
	i3 = root.ptm.values().iterator();
	i4 = root.pts.iterator();
	
	for (i = 0; i < nNodes; i++) { 
	    entry = (Map.Entry)i1.next();
	    Assert.that(i2.next() == entry.getKey());
	    Assert.that(i3.next() == entry.getValue());
	    Assert.that(i4.next().equals(entry.getValue()));
	    Assert.that(entry.getKey().equals("Key" + key[i]));
	    Assert.that(((DbRecord)entry.getValue()).value.equals("Rec" + key[i]));
	    i4.remove();
	}
	Assert.that(!i1.hasNext());
	Assert.that(!i2.hasNext());
	Assert.that(!i3.hasNext());
	Assert.that(!i4.hasNext());
	
	System.out.println("Tree iterators test passed");

	for (i = 0; i < nNodes; i++) { 
	    rec = (DbRecord)root.ptm.remove("Key" + key[i]);
	    Assert.that(rec != null);
	    Assert.that(rec == root.phm.remove("Key" + key[i]));
	}
	root.pal.clear();
	System.out.println("Cleanup completed");	
    }


    void runTestTransaction() { 
	int i;
	DbRecord rec;
	Map.Entry entry;

	String key[] = new String[nNodes];
	Random rnd = new Random(2001);
	for (i = 0; i < nNodes; i++) { 
	    key[i] = Long.toString(rnd.nextLong());
	}
	
	for (i = 0; i < nNodes; i++) { 
	    insert(key[i]);
	}
	System.out.println("Database initialized");

	Assert.that(pal.size() == nNodes);
	Assert.that(phm.size() == nNodes);
	Assert.that(ptm.size() == nNodes);
	Assert.that(phs.size() == nNodes);
	Assert.that(pts.size() == nNodes);
	
	for (i = 0; i < nNodes; i++) { 
	    rec = (DbRecord)pal.get(i);
	    Assert.that(rec.value.equals("Rec" + key[i]));
	    Assert.that(phm.get("Key" + key[i]) == rec);
	    Assert.that(ptm.get("Key" + key[i]) == rec);
	    Assert.that(phs.contains(rec));
	    Assert.that(pts.contains(rec));
	}	
	System.out.println("Random search test passed");
	
	Arrays.sort(key);
	
	Iterator i1 = phm.entrySet().iterator();
	Iterator i2 = phm.keySet().iterator();
	Iterator i3 = phm.values().iterator();
	Iterator i4 = phs.iterator();
	boolean mark[] = new boolean[key.length];
	
	for (i = 0; i < nNodes; i++) { 
	    entry = (Map.Entry)i1.next();
	    Assert.that(i2.next() == entry.getKey());
	    Assert.that(i3.next() == entry.getValue());
	    String k = entry.getKey().toString().substring(3);
	    int j = Arrays.binarySearch(key, k);
	    Assert.that(j >= 0);
	    Assert.that(!mark[j]);
	    Assert.that(entry.getKey().equals("Key" + key[j]));
	    Assert.that(((DbRecord)entry.getValue()).value.equals("Rec" + key[j]));
	    Assert.that(i4.next() != null);
	    mark[j] = true;
	    i4.remove();
	}
	Assert.that(!i1.hasNext());
	Assert.that(!i2.hasNext());
	Assert.that(!i3.hasNext());
	Assert.that(!i4.hasNext());
	System.out.println("Hash iterators test passed");
	
	i1 = ptm.entrySet().iterator();
	i2 = ptm.keySet().iterator();
	i3 = ptm.values().iterator();
	i4 = pts.iterator();
	
	for (i = 0; i < nNodes; i++) { 
	    entry = (Map.Entry)i1.next();
	    Assert.that(i2.next() == entry.getKey());
	    Assert.that(i3.next() == entry.getValue());
	    Assert.that(i4.next().equals(entry.getValue()));
	    Assert.that(entry.getKey().equals("Key" + key[i]));
	    Assert.that(((DbRecord)entry.getValue()).value.equals("Rec" + key[i]));
	    i4.remove();
	}
	Assert.that(!i1.hasNext());
	Assert.that(!i2.hasNext());
	Assert.that(!i3.hasNext());
	Assert.that(!i4.hasNext());
	
	System.out.println("Tree iterators test passed");

	for (i = 0; i < nNodes; i++) { 
	    rec = (DbRecord)ptm.remove("Key" + key[i]);
	    Assert.that(rec != null);
	    Assert.that(rec == phm.remove("Key" + key[i]));
	}
	pal.clear();
	System.out.println("Cleanup completed");	
    }
}
		


