import goodsjpi.*;


class L2ListNode extends Persistent { 
    L2ListNode next;
    L2ListNode prev;
    int        value;

    L2ListNode() { 
	next = prev = this;
    }

    void increment() { 
	value += 1;
    }

    void decrement() { 
	value -= 1; 
    }

    void update() { 
	increment();
	next.decrement();
    }

    void link(L2ListNode node) {
	node.next = this;
	node.prev = prev;
	prev = prev.next = node;
    }

    int calculateSum() { 
	L2ListNode node = this; 
	int sum = 0;
	do { 
	    sum += node.value;
	} while ((node = node.next) != this);
	return sum;
    }
    
    void updateNode(int index) { 
	L2ListNode node = next; 
	while (--index >= 0) { 
	    node = node.next;
	}
	node.update();
    }	

}

class TestThread extends Thread { 
    public void run() { 
	Database db = new Database();
	db.attach();
	try { 
	    if (db.open("testtrans.cfg")) {
		L2ListNode list = (L2ListNode)db.getRoot();
		L2ListNode node;
		int i, j; 
		if (list == null) {
		    list = new L2ListNode();
		    for (i = 0; i < TestTrans.nNodes; i++) {
			node = new L2ListNode();
			node.value = i+1;
			list.link(node);
		    }
		    db.setRoot(list);
		}
		for (i = 0; i < TestTrans.nIterations; i++) { 
		    int sum = list.calculateSum();
		    if (sum != TestTrans.nNodes*(TestTrans.nNodes+1)/2) { 
			System.err.println("Wrong check sum: " + sum + " instead of "  
					   + TestTrans.nNodes*(TestTrans.nNodes+1)/2);
		    }
		    list.updateNode(i % TestTrans.nNodes);
		}
		db.close();
	    } else { 
		System.err.println("Thread failed to open database");
	    }
	} finally { 
	    db.detach();
	}
    }
}
 
public class TestTrans { 
    static final int nNodes = 100;
    static final int nIterations = 100;
    static final int nThreadsByDefault = 10;

    static public void main(String args[]) throws Exception { 
	int nThreads = nThreadsByDefault;
	int i;
	if (args.length > 0) { 
	    nThreads = Integer.parseInt(args[0], 10);
	}
	Thread[] threads = new Thread[nThreads];
	for (i = 0; i < nThreads; i++) { 
	    threads[i] = new TestThread();
	    threads[i].start();
	}
	for (i = 0; i < nThreads; i++) { 
	    threads[i].join();
	}
	Database db = new Database();
	if (db.open("testtrans.cfg")) {
	    L2ListNode list = (L2ListNode)db.getRoot();
	    if (list.value != -nIterations/nNodes*nThreads || list.next.value != nIterations/nNodes*nThreads+1) { 
		System.err.println("Something is definitly wrong: list.value=" 
				   + list.value + " instead of " + -nIterations/nNodes*nThreads 
				   + ", list.next.value=" + list.next.value + " instead of " 
				   + nIterations/nNodes*nThreads+1);
	    } else { 
		System.out.println("Test successfully passed");
	    }
	    list.value = 0;
	    list.next.value = 1;
	    db.close();
	} else {
	    System.err.println("Failed to open detabase");
	}
    }
}





