/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.CharacterData;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.lang.StringIndexOutOfBoundsException;

/**
 * PersistentCharacterData implementation of w3c CharacterData
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public abstract class PersistentCharacterData extends PersistentNode implements CharacterData 
{
   /** Empty child nodes. */
   private static transient NodeList singletonNodeList = new NodeList() 
   {
      public Node item(int index)
      {
         return null;
      }
      public int getLength()
      {
         return 0;
      }
   };

   public PersistentCharacterData(Document doc, short type, String name)
   {
      super(doc, type, name);
   }


   /** Returns the node name. */
   public abstract String getNodeName();

   /** Returns an empty node list. */
   public NodeList getChildNodes() 
   {
      return singletonNodeList;
   }

   public String getData() throws DOMException
   {
      return value;
   }

   public void setData(String data) throws DOMException
   {
      value = data;
   }

   public int getLength()
   {
      return value.length();
   }

   public String substringData(int offset, int count) throws DOMException
   {
      int length = value.length();
      if(count < 0 || offset < 0 || offset > length - 1)
      {
         throw new PersistentDOMException(DOMException.INDEX_SIZE_ERR, 
                                    "DOM004 Index out of bounds");
      }

      int tailIndex = Math.min(offset + count, length);

      return value.substring(offset, tailIndex);
   }

   public void appendData(String arg)
   {
      setNodeValue(value + arg);
   }

   public void insertData(int offset, String arg) throws DOMException
   {
      try
      {
         value = new StringBuffer(value).insert(offset, arg).toString();
      }
      catch(StringIndexOutOfBoundsException e)
      {
         throw new PersistentDOMException(DOMException.INDEX_SIZE_ERR, 
                                    "DOM004 Index out of bounds");
      }
   }

   public void deleteData(int offset, int count) throws DOMException
   {
      if(count < 0)
      {
         throw new PersistentDOMException(DOMException.INDEX_SIZE_ERR, 
                                    "DOM004 Index out of bounds");
      }

      int tailLength = Math.max(value.length() - count - offset, 0);
      try
      {
         value = value.substring(0, offset) + (tailLength > 0 
                                               ? value.substring(offset + count, offset + count + tailLength) 
                                               : "");

      }
      catch(StringIndexOutOfBoundsException e)
      {
         throw new PersistentDOMException(DOMException.INDEX_SIZE_ERR, 
                                    "DOM004 Index out of bounds");
      }
   }

   public void replaceData(int offset, int count, String arg) throws DOMException
   {
      deleteData(offset, count);
      insertData(offset, arg);
   }

   public String toString()
   {
      return value;
   }
}

