/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.DOMException;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Attr;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NodeList;
import org.w3c.dom.DocumentType;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.w3c.dom.Comment;
import org.w3c.dom.CDATASection;

import goodsjpi.Database;
import goodsjpi.BasicMetaobject;


/**
 * PersistentDocument implementation of w3c Document
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public class PersistentDocument extends PersistentNode implements Document 
{
   private static String      NODE_NAME = "#document";

   public PersistentDocument()
   {
      super(null, Node.DOCUMENT_NODE, NODE_NAME);
      super.setDocument(this);
   }

   public static Document getRoot(Database db)
   {

      PersistentDocument root;

      root = (PersistentDocument) db.getRoot();

      if(root == null)
      {

         BasicMetaobject.beginNestedTransaction ();
         System.out.println("Initialize Database for the first time");

         root = new PersistentDocument();
         db.setRoot(root);
         BasicMetaobject.endNestedTransaction();
      }


      System.out.println("getRoot exit");
      return(Document) root;      
   }   


   public DocumentType getDoctype()
   {
      return null; // not supported
   }

   public DOMImplementation  getImplementation()
   {
      return null; // TODO
   }

   public Element getDocumentElement()
   {
      Element element = (Element)getFirstChild();
      while(element != null && element.getNodeType() != Node.ELEMENT_NODE)
         element = (Element)element.getNextSibling();
      return element;
   }

   public Document getDocument ()
   {
      return this;
   }

   public String getNodeName ()
   {
      return NODE_NAME; // return a default constant name.
   }

   public Element createElement(String tagName) throws DOMException
   {
      return new PersistentElement(this, tagName);
   }

   public DocumentFragment createDocumentFragment()
   {
      return new PersistentDocumentFragment(this, "#DocumentFragment");
   }

   public Text createTextNode(String data)
   {
      PersistentText monTexte = new PersistentText(this, "#text");
      monTexte.setData(data);
      return monTexte;
   }

   public Comment createComment(String data)
   {
      PersistentComment monCommentaire = new PersistentComment(this, "#Comment");
      monCommentaire.setData (data);

      return monCommentaire;
   }

   public CDATASection createCDATASection(String data) throws DOMException
   { 
      PersistentCDATASection pcs = new PersistentCDATASection(this, "#CDATA");
      pcs.setData(data);
      return pcs;
   }

   public ProcessingInstruction createProcessingInstruction(String target, String data) throws DOMException
   {
      
      PersistentProcessingInstruction ppi = new PersistentProcessingInstruction(this, "#ProcessingInstruction");
      ppi.setData (data);
      ppi.setTarget (target);
      return ppi;
   }

   public Attr createAttribute(String name) throws DOMException
   {
      return new PersistentAttr(this, name);
   }

   public EntityReference createEntityReference(String name) throws DOMException
   {
      return new PersistentEntityReference (this, name);
   }

   public NodeList getElementsByTagName(String tagname)
   {
      return null; // not supported... yet !
   }

}

