/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.DOMException;

/**
 * PersistentElement implementation of w3c Element
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public class PersistentElement extends PersistentNode implements Element 
{
   private PersistentNamedNodeMap attributes;
   public static final String EMPTY_STRING="";

   public PersistentElement(Document associatedDocument, String name)
   {
      super(associatedDocument, ELEMENT_NODE, name);
      this.name=name;
   }

   public Element getChildByTagName (String tagName)
   {
      tagName = tagName.intern();
      NodeList childList = getChildNodes();
      for(int i = 0; i < childList.getLength(); i++)
      {
         Node child = childList.item(i);
         if(child != null && child.getNodeType() == Node.ELEMENT_NODE &&
            tagName == child.getNodeName())
            return(Element)child;
      }
      return null;
   }


   public Element getAncestorByTagName (String tagName)
   {
      Node node = this;
      while((node = node.getParentNode()) != null)
         if(node.getNodeType() == Node.ELEMENT_NODE &&
            tagName == node.getNodeName())
            return(Element)node;

      return null;
   }



   public String toString ()
   {
      StringBuffer buf = new StringBuffer();

      buf.append('<');

      
      buf.append(name);
      
      NamedNodeMap attrs = getAttributes();
      
      if(attrs != null)
      {
         for(int i = 0; i < attrs.getLength(); i++)
         {
            buf.append(' ');
            buf.append(attrs.item(i));
         }

      }
      buf.append(">\n");

      buf.append(super.toString());

      buf.append("</");
      
      buf.append(name);
     
      buf.append(">\n");

      return buf.toString();
   }

   // override to inheritate the property
   public void setDocument (Document doc)
   {
      super.setDocument(doc);

      PersistentNamedNodeMap list = attributes;
      if(list != null && list.getLength() > 0)
      {
         for(int i = 0; i < list.getLength(); i++)
         {
            Node attr = list.item(i);
            if(attr != null)
               ((PersistentNode)attr).setDocument(doc);
         }
      }
   }

   public String getAttribute(String name)
   {
      if(attributes == null)
         return EMPTY_STRING;

      Node attr = attributes.getNamedItem(name);
      return attr != null ? attr.getNodeValue() : EMPTY_STRING;
   }

   public String getTagName()
   {
      return super.getNodeName();
   }

   public void setAttribute(String name, String value)
   {
      if(attributes == null)
         createAttributeList();

      Node attr = attributes.getNamedItem(name);
      if(attr == null)
      {
         attr = getOwnerDocument().createAttribute(name);
         attributes.setNamedItem(attr);
      }
      attr.setNodeValue(value);
   }

   public void removeAttribute(String name)
   {
      if(attributes != null)
         attributes.removeNamedItem(name);
   }

   public Attr getAttributeNode(String name)
   {
      return attributes != null ? (Attr)attributes.getNamedItem(name) : null;
   }

   public Attr setAttributeNode(Attr newAttr) throws DOMException
   {
      Attr oldAttr = null;
      if(attributes != null)
         oldAttr = (Attr)attributes.getNamedItem(newAttr.getNodeName());
      else
         createAttributeList();

      attributes.setNamedItem(newAttr);
      return oldAttr;
   }

   public Attr removeAttributeNode(Attr oldAttr) throws DOMException
   {
/*      if (attributes == null)
         throw new PersistentDOMException(DOMException.NOT_FOUND_ERR,"");

      Attr attr = (Attr)attributes.removeNamedItem(oldAttr.getNodeName());
      if (attr == null || attr != oldAttr)
         throw new PersistentDOMException(DOMException.NOT_FOUND_ERR,"");
*/
      //TODO
      return oldAttr;
   }


   public NodeList getElementsByTagName(String name)
   {
      /*  if ("*".equals(name))
            return DOMUtil.getNodesByFilter(this, BasicElement.NODETYPE_FILTER);
         else
            return DOMUtil.getNodesByFilter(this, new TagNameFilter(name));
            */
      // TODO
      return null;
   }

   public void normalize()
   {
      if(hasChildNodes())
      {
         Node node = getFirstChild();
         while(node != null)
         {
            if(node.getNodeType() == Node.TEXT_NODE)
            {
               Text text = (Text)node;
               Node next;
               while((next = text.getNextSibling()) != null &&
                     next.getNodeType() == TEXT_NODE)
               {
                  text.appendData(next.getNodeValue());
                  getParentNode().removeChild(next);
               }
            }
            else if(node.getNodeType() == Node.ELEMENT_NODE)
               ((Element)node).normalize();

            node = node.getNextSibling();
         }
      }
   }

   protected PersistentNamedNodeMap  createAttributeList()
   {
      if(attributes == null)
         attributes = new PersistentNamedNodeMap(false);
      return attributes;
   }

}

