/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;

/**
 * PersistentNamedNodeMap implementation of w3c NamedNodeMap
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public class PersistentNamedNodeMap extends PersistentNodeList implements NamedNodeMap 
{
   private boolean      ignoreCase;

   public PersistentNamedNodeMap ()
   {
      this.ignoreCase = false;
   }

   public PersistentNamedNodeMap (boolean ignoreCase)
   {
      this.ignoreCase = ignoreCase;
   }

   public PersistentNamedNodeMap (PersistentNamedNodeMap original, boolean deep)
   {
      super(original.getLength());

      ignoreCase = original.ignoreCase;

      for (int i = 0; i < original.getLength(); i++)
         appendNode(original.item(i).cloneNode(deep));
   }

   public boolean getIgnoreCase ()
   {
      return ignoreCase;
   }

   public void setIgnoreCase (boolean ignoreCase)
   {
      this.ignoreCase = ignoreCase;
   }

   public int getNamedIndex (String name)
   {
      if (ignoreCase)
      {
         for (int i = 0; i < getLength(); i++)
         {
            Node node = item(i);
            if (name.equalsIgnoreCase(node.getNodeName()))
               return i;
         }
      }
      else
      {
         for (int i = 0; i < getLength(); i++)
         {
            Node node = item(i);
            if (name.equals(node.getNodeName()))
               return i;
         }
      }
      return -1;
   }

// org.w3c.dom.NamedNodeMap

   public Node getNamedItem(String name)
   {
      int index = getNamedIndex(name);
      if (index >= 0)
         return item(index);
      else
         return null;
   }

   public Node setNamedItem(Node node)
   {
      int index = getNamedIndex(node.getNodeName());
      if (index >= 0)
      {
         Node oldNode = item(index);
         replaceNode(node, index);
         return oldNode;
      }
      else
      {
         appendNode(node);
         return null;
      }
   }

   public Node removeNamedItem(String name)
   {
      int index = getNamedIndex(name);
      if (index >= 0)
      {
         Node oldNode = item(index);
         removeNode(index);
         return oldNode;
      }
      else
         return null;
   }
}

