/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.Node;
import org.w3c.dom.DOMException;
import org.w3c.dom.NodeList;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import goodsjpi.Persistent;


/**
 * PersistentNode implementation of w3c Node
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public class PersistentNode extends Persistent implements Node 
{
   protected PersistentNode parent;
   protected String value;
   protected String name;
   private short type;
   protected PersistentNodeList children;
   private PersistentNamedNodeMap attributes;
   protected PersistentDocument associatedDocument;

   public PersistentNode(Document associatedDocument, short type, String name)
   {
      this.name=name;
      this.type=type;
      this.associatedDocument=(PersistentDocument) associatedDocument;
      children=new PersistentNodeList();
   }

   public void setParentNode(PersistentNode parent)
   {
      this.parent=parent;
   }

   public void setDocument(Document newAssociatedDocument)
   {
      this.associatedDocument=(PersistentDocument) newAssociatedDocument;
   }


   public String getNodeName()
   {
      return name;
   }
   public String getNodeValue() throws DOMException
   {
      return value;
   }

   public void setNodeValue(String nodeValue) throws DOMException
   {
      value=nodeValue;
   }

   public short getNodeType()
   {
      return type;
   }

   public Node getParentNode()
   {
      return parent;
   }

   public NodeList getChildNodes()
   {
      return children;
   }

   public Node getFirstChild()
   {
      return children.getFirstNode();
   }

   public Node getLastChild()
   {
      return children.getLastNode();
   }

   public Node getPreviousSibling()
   {
      if(parent!=null)
         return parent.children.getPreviousNode(this);
      else
         return null;
   }

   public Node getNextSibling()
   {
      if(parent!=null)
         return parent.children.getNextNode(this);
      else
         return null;
   }

   public NamedNodeMap getAttributes()
   {
      return attributes;
   }

   public Document getOwnerDocument()
   {
      return associatedDocument;
   }

   public Node insertBefore(Node newChild, Node refChild) throws DOMException
   {
      children.insertNode(newChild, children.getNodeIndex(refChild));
      ((PersistentNode) newChild).setParentNode(this);
      return newChild;
   }

   public Node replaceChild(Node newChild, Node oldChild) throws DOMException
   {
      children.replaceNode(newChild,children.getNodeIndex(oldChild));
      ((PersistentNode)newChild).setParentNode(this);
      return newChild;
   }

   public Node removeChild(Node oldChild) throws DOMException
   {
      children.removeNode(oldChild);
      return oldChild;
   }

   public Node appendChild(Node newChild) throws DOMException
   {
      children.insertNode(newChild,children.getLength());
      ((PersistentNode)newChild).setParentNode(this);
      return newChild;
   }

   public boolean hasChildNodes()
   {
      return children.getLength()!=0;
   }

   public Node cloneNode(boolean deep)
   {
      return null; // TODO
   }

   public String toString()
   {
      StringBuffer answer= new StringBuffer();

      Node currentNode = getFirstChild();

      while(currentNode != null)
      {
         answer.append(currentNode.toString());
         currentNode = currentNode.getNextSibling();
      }
      return new String(answer);

   }

}

