/*
 * $Id: $
 */

package com.neurothemis.xmlswarm.parser;

import gd.xml.*;
import org.w3c.dom.*;
import java.util.Hashtable;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Enumeration;
import goodsjpi.Persistent;

/**
 * XML parser.
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */

public class Parser extends Persistent implements XMLResponder
{
   private static final boolean DEBUG=true;

   private transient Document doc;
   private transient Node currentNode;
   private transient InputStream xmlInput;
   private transient String prefix;


   public Parser()
   {
   }

   public synchronized void parse(InputStream xmlInput, Document destinationDocument) throws ParseException
   {
      this.xmlInput=xmlInput;
      this.doc=destinationDocument;

      if (DEBUG)
         prefix = "> ";

      XMLParser xp = new XMLParser();
      xp.parseXML(this);
      
   }

   /* DTD METHODS */

   public void recordNotationDeclaration(String name, String pubID, String sysID) throws ParseException
   {
      if (DEBUG)
      {
         System.out.print(prefix+"!NOTATION: "+name);
         if (pubID!=null) System.out.print("  pubID = "+pubID);
         if (sysID!=null) System.out.print("  sysID = "+sysID);
         System.out.println("");
      }
   }

   public void recordEntityDeclaration(String name, String value, String pubID, String sysID, String notation) throws ParseException
   {
      if (DEBUG)
      {
         System.out.print(prefix+"!ENTITY: "+name);
         if (value!=null) System.out.print("  value = "+value);
         if (pubID!=null) System.out.print("  pubID = "+pubID);
         if (sysID!=null) System.out.print("  sysID = "+sysID);
         if (notation!=null) System.out.print("  notation = "+notation);
         System.out.println("");
      }
   }

   public void recordElementDeclaration(String name, String content) throws ParseException
   {
      if (DEBUG)
      {
         System.out.print(prefix+"!ELEMENT: "+name);
         System.out.println("  content = "+content);
      }
   }

   public void recordAttlistDeclaration(String element, String attr, boolean notation, String type, String defmod, String def) throws ParseException
   {
      if (DEBUG)
      {
         System.out.print(prefix+"!ATTLIST: "+element);
         System.out.print("  attr = "+attr);
         System.out.print("  type = " + ((notation) ? "NOTATIONS " : "") + type);
         System.out.print("  def. modifier = "+defmod);
         System.out.println( (def==null) ? "" : "  def = "+notation);
      }
   }

   public void recordDoctypeDeclaration(String name, String pubID, String sysID) throws ParseException
   {
      if (DEBUG)
      {
         System.out.print(prefix+"!DOCTYPE: "+name);
         if (pubID!=null) System.out.print("  pubID = "+pubID);
         if (sysID!=null) System.out.print("  sysID = "+sysID);
         System.out.println("");
         prefix = "";
      }
   }


    /* DOC METHODS */

   public void recordDocStart()
   {
      if (DEBUG)
         System.out.println("Doc Start");
      currentNode=doc;
   }

   public void recordDocEnd()
   {
      if (DEBUG)
      {
         System.out.println("");
         System.out.println("Parsing finished without error");
      }
   }

   public void recordElementStart(String name, Hashtable attr) throws ParseException
   {
      System.out.println(prefix+"Element: "+name);
      Element element = doc.createElement (name);
      
      if (attr!=null)
      {
         String conj = "";
         Enumeration e = attr.keys();
	      if (DEBUG)
         {
            System.out.print(prefix);
         }

         while (e.hasMoreElements())
         {
            Object k = e.nextElement();

            if (DEBUG)
            {
               System.out.print(conj+k+" = "+attr.get(k));
               conj = ", ";
            }

            element.setAttribute(k.toString(), attr.get(k).toString());
         }

         if (DEBUG)
            System.out.println("");
      }
      if (DEBUG)
         prefix = prefix+"  ";
      currentNode.appendChild(element);
      currentNode=element;
   }

   public void recordElementEnd(String name) throws ParseException
   {
      if (DEBUG)
         prefix = prefix.substring(2);
      currentNode=currentNode.getParentNode();
   }

   public void recordPI(String name, String pValue)
   {
      if (DEBUG)
         System.out.println(prefix+"*"+name+" PI: "+pValue);
   }

   public void recordCharData(String charData)
   {
      if (DEBUG)
         System.out.println(prefix+charData);
     
     Text textNodeToInsert=doc.createTextNode(charData); 
     currentNode.appendChild(textNodeToInsert);
   }

   public void recordComment(String comment)
   {
      if (DEBUG)
         System.out.println(prefix+"*Comment: "+comment);
   }

    /* INPUT METHODS */

   public InputStream getDocumentStream() throws ParseException
   {
      return xmlInput;
   }

   public InputStream resolveExternalEntity(String name, String pubID, String sysID) throws ParseException
   {
      return null;
      /*if (sysID!=null)
      {
         File f = new File((new File(filename)).getParent(), sysID);
         try
         {
            return new FileInputStream(f);
         }
         catch (FileNotFoundException e)
         {
            throw new ParseException("file not found ("+f+")");
         }
      }
      else return null;*/
   }

   public InputStream resolveDTDEntity(String name, String pubID, String sysID) throws ParseException
   {
      return resolveExternalEntity(name, pubID, sysID);
   }
}
