//-< BasicMetaobject.java >------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 18-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Basic method methods implementation
//-------------------------------------------------------------------*--------*

package goodsjpi;

public abstract class BasicMetaobject extends Metaobject { 
    //
    // Invocation of this method is inserted by MOP geneator before each
    // method invocation or object component access.
    //
    public void preDaemon(Object o, int attr) {
	Persistent obj = (Persistent)o;
	CacheManager mng  = CacheManager.getCacheManager();
	synchronized(obj) {
	    if (obj.accessCount == 0) { 
		synchronized (this) { 
		    obj.accessCount = 1;
		    //
		    // Setting of access count to non-zero prevents from
		    // object destruction by receiving deterioration 
		    // notification from server. Instead of this "invalid"
		    // object attribute will set to true.
		    //
		}
		synchronized (mng.globalCS) { 
		    mng.nestedTransactionCount += 1;
		    if (obj.opid != 0) { // persistent object
			mng.removeFromCache(obj);
		    }
		}
		if (obj.opid != 0) { // presistent objects
		    if ((attr & Metaobject.MUTATOR) != 0){
			beginWriteAccess(obj);
		    } else { 
			beginReadAccess(obj);
		    }
		    if (obj.invalidated) { 
			//
			// In case of using pessimistic scheme object is 
			// already locked at this moment, so it can not be 
			// more changed by another application before the end 
			// of transaction. 
			// And with optimistic scheme... hmmm. it nice to be 
			// optimist.
			//
			synchronized (this) { 
			    obj.invalidated = false;
			    destroyObject(obj);
			}
		    }
		    if ((obj.state & Persistent.DESTRUCTED) != 0) { 
			obj.storage.load(obj.opid, obj);
		    } else if ((obj.state & Persistent.RAW) != 0) {
			obj.storage.prepare(obj);
		    }
		}
	    } else { 
		obj.accessCount += 1;
	    }
	}
    }
    
    //
    // Invocation of this method is inserted by MOP geneator after each
    // method invocation or object component access.
    //
    public void postDaemon(Object o, int attr, boolean modified) {
	Persistent obj = (Persistent)o;
	CacheManager mng  = CacheManager.getCacheManager();
	synchronized(obj) {
	    if (modified) { 
		obj.state |= Persistent.DIRTY;
	    }
	    if (obj.accessCount == 1) { 
		if (obj.opid != 0) { 
		    endAccess(obj);
		}
		synchronized (mng.globalCS) { 
		    if (obj.opid != 0) { // persistent object
			if ((obj.state & (Persistent.NOTIFY|
			     Persistent.TRANSREAD|Persistent.TRANSWRITE)) == 0)
			{
			    mng.insertInCache(obj);
			}
		    }
		    if (--mng.nestedTransactionCount == 0) { 
			if ((attr & Metaobject.EXCEPTION) != 0) {
			    abortTransaction();
			} else { 
			    commitTransaction();
			}
		    }		
		}
		synchronized (this) { 
		    //
		    // We delay decrement of access counter till this moment
		    // to prevent object destruction on receiving deterioration
		    // notification from server. Synchronized construction 
		    // for metaobject object provides mutual exclusion with 
		    // objectInvalidate() method.
		    //
		    obj.accessCount = 0;
		    if (obj.invalidated) { 
			obj.invalidated = false;
			synchronized (mng) { 
			    mng.removeFromCache(obj);
			    destroyObject(obj);
			}
		    }
		}
	    } else { 
		obj.accessCount -= 1;
	    }
	}
    }		

    //
    // Assign persistent identifier to the transient object of persistent
    // capable class. The object is placed in the same storage as 
    // persistent object containing reference to this object unless
    // explicit storage attachment was specified for this object.
    //
    protected void makePersistent(Persistent obj, Persistent parent)
    {
	if (obj.storage == null) { 
	    //
	    // By default transient object is placed in the same storage as 
	    // persitent object referenced this object.
	    //
	    obj.storage = parent.storage;
	} else { 
	    // 
	    // Interdatabase references are prohibited
	    //
	    Assert.that(obj.storage.database == parent.storage.database);
	}
	obj.storage.allocate(obj, 0);
	//
	// Object is linked in transaction list after persistent object, 
	// from which it is referenced, so this object will handled by
	// commitTransaction() method soon after it complets with parent 
	// object.
	//
	CacheManager mng  = CacheManager.getCacheManager();
	mng.linkAfter(parent, obj); 
	obj.state |= Persistent.TRANSWRITE|Persistent.DIRTY;
    }

    //
    // Object thrown by GC from weak cache are linked into forgottenObjectList
    // list and notifications are send to servers at the transaction
    // commit time. This is done because clearing weak reference by GC 
    // is not saynchronized with invocation of finalize() method. So there is
    // no warrenty that in the moment between object is removed from weak cache
    // and ionvocation of finilize method, the object with the same identifier
    // will not be retirieved from the server. Delayed strategy of sending
    // forget message together with checking object for presence in the cache 
    // before sending foget message to the server makes sending of wrong
    // forgot message not possible.
    // 
    protected static void sendForgottenObjectsToServers(CacheManager mng) { 
	synchronized (mng.forgottenObjectList) {
	    Persistent obj;
	    Persistent list = mng.forgottenObjectList;
	    while (list.next != list) { 
		Storage store = list.next.storage;
		int n = 0;
		for (obj = list.next; obj != list; obj = obj.next) {
		    if (obj.storage == store) { 
			Object o = store.objectCache.get(obj.opid);
			if (o == obj) { 
			    store.objectCache.remove(obj.opid);
			    n += 1;
			} else if (o == null) { 
			    n += 1;
			} else { 
			    obj.next.prev = obj.prev;
			    obj.prev.next = obj.next;
			}
		    }
		}
		if (n > 0) { 
		    int[] opids = new int[n];
		    for (obj = list.next; obj != list; obj = obj.next) {
			if (obj.storage == store) { 
			    opids[--n] = obj.opid;
			    obj.next.prev = obj.prev;
			    obj.prev.next = obj.next;
			}
		    }
		    store.forgetObjects(opids);
		}
	    }
	}
    }
	    
    //
    // At this moment there are no more active methods for 
    // classes controlled by this metaobject (or metaobjects dericed from it).
    // This method will try to commit all transactions in all opened databases.
    // Until this method is completed no other invocation or access to
    // objects controled by this MOP is possoble (it is forced by declaring
    // this static method "synchronized" as well as static method fixObject(),
    // which is called before any access to the object).
    //
    protected static void commitTransaction() {	
	Persistent next;
	CacheManager mng  = CacheManager.getCacheManager();
	sendForgottenObjectsToServers(mng);
	Persistent transObjects = mng.transObjects;
      loop:
	while (transObjects.next != transObjects) { 
	    Persistent obj = transObjects.next;
	    if ((obj.state & Persistent.DIRTY) == 0) { 
		// 
		// As far as all modified objects are placed at the head of 
		// the list, and we are navigating through the list startig 
		// from the head, finding of non-modified object means that
		// there are no more other modified objects. So no more 
		// other transactions has to be send to the server. 
		// We only should release all locks hold by these transactions.
		//
		do { 
		    next = obj.next;
		    obj.metaobject.releaseObject(obj);
		} while((obj = next) != transObjects);
		break;
	    }
	    int nTransServers = 0;
	    Storage coordinator = obj.storage;
	    Database db = coordinator.database;
	    Storage[] servers = new Storage[db.storages.length];

	    do { // find all tranasction objects from the same database
		Storage storage = obj.storage;
		if (storage.database == db) { 
		    if (obj.invalidated) { 
			abortTransaction(mng, db);
			continue loop;
		    }
		    if (!storage.isInvolvedInTransaction()) { 
			if (storage.id < coordinator.id) {
			    coordinator = storage;
			}
			servers[nTransServers++] = storage;
		    }
		    storage.store(obj);
		}
		obj = obj.next;
	    } while (obj != transObjects);

	    if (nTransServers != 0) { 
		//
		// commit local transaction or coordinator part of global one
		//
		int tid = coordinator.commitTransaction(nTransServers,servers);
		if (tid == -1) { 
		    abortTransaction(mng, db);
		    continue;
		}
		if (nTransServers > 1) { // global transaction
		    for (int i = 0; i < nTransServers; i++) { 
			if (servers[i] != coordinator) { 
			    servers[i].commitSubtransaction(coordinator,
							    nTransServers, 
							    servers, tid);
			}
		    }
		    if (!coordinator.waitTransactionCompletion()) { 
			abortTransaction(mng, db);
			continue;
		    }
		}
	    }
	    //
	    // Transaction was successfully commited
	    //
	    for (obj = transObjects.next; obj != transObjects; obj = next){
		next = obj.next;
		if (obj.storage.database == db) { 
		    obj.metaobject.commitObjectChanges(obj);
		}
	    }
	}
    }
    
    protected static void abortTransaction(CacheManager mng, Database db) { 
        Persistent obj, next;

	Persistent transObjects = mng.transObjects;
	for (obj = transObjects.next; obj != transObjects; obj = next){
	    next = obj.next;
	    if (obj.storage.database == db) { 
	        obj.metaobject.undoObjectChanges(obj);
	    }
	}
	db.abortTransaction();
    }

    protected static void abortTransaction() { 
	CacheManager mng = CacheManager.getCacheManager();
	Persistent transObjects = mng.transObjects; 
	while (transObjects.next != transObjects) { 
	    abortTransaction(mng, transObjects.next.storage.database);
	}
    }

    protected synchronized void commitObjectChanges(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
	Assert.that((obj.state & (Persistent.TRANSREAD|Persistent.TRANSWRITE))
		    != 0);
	CacheManager.unlink(obj); // remove object from transaction list
	obj.state &= ~(Persistent.TRANSREAD | Persistent.TRANSWRITE |
		       Persistent.XLOCKED | Persistent.SLOCKED | 
		       Persistent.DIRTY | Persistent.NEW);
	if (obj.invalidated) { 
	    obj.invalidated = false;
	    destroyObject(obj);
	} else if ((obj.state & Persistent.NOTIFY) == 0) { 
	    mng.insertInCache(obj);
	}
    }
	
    protected synchronized void undoObjectChanges(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
	Assert.that((obj.state & (Persistent.TRANSREAD|Persistent.TRANSWRITE))
		    != 0);
	CacheManager.unlink(obj); // remove object from transaction list

	if ((obj.state & (Persistent.SLOCKED|Persistent.XLOCKED)) != 0) {
	    obj.storage.unlock(obj, Protocol.lck_none);
	}
	
	if ((obj.state & Persistent.DIRTY) != 0) {
	    if ((obj.state & Persistent.NEW) == 0) {
		//
		// All persistent objects should be reloaded from the server.
		//
		destroyObject(obj);
	    } else { 
		obj.storage.forgetObject(obj);
		obj.opid = 0; // object beacomes transsient once again
		obj.storage = null;
	    }
	} else if ((obj.state & Persistent.NOTIFY) == 0) {
	    mng.insertInCache(obj);
	}
	obj.state &= ~(Persistent.TRANSREAD | Persistent.TRANSWRITE |
		       Persistent.XLOCKED | Persistent.SLOCKED | 
		       Persistent.DIRTY | Persistent.NEW);
    }
	
    protected void preloadObject(Persistent obj, int size) { 
	CacheManager mng = CacheManager.getCacheManager();
        synchronized (mng) { 
	    if (obj.data != null) { 
		mng.cacheSize0 -= obj.data.length;
	    } else { 
		mng.cacheSize0 += CacheManager.objectAllocationOverhead + obj.desc.fixedSize;
	    }
	    mng.cacheSize0 += size;
	}
    }

    protected void loadObject(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
        synchronized (mng) {
	    mng.cacheSize0 -= CacheManager.objectAllocationOverhead + obj.data.length + obj.desc.fixedSize;
	}
    }

    protected void forgetObject(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
        synchronized (mng) { 
	    if (obj.data != null) { 
		mng.cacheSize0 -= CacheManager.objectAllocationOverhead + obj.data.length + obj.desc.fixedSize;
		obj.data = null;
	    }
	}
	synchronized (mng.forgottenObjectList) {
	    obj.next = mng.forgottenObjectList.next;
	    obj.prev = mng.forgottenObjectList;
	    obj.next.prev = mng.forgottenObjectList.next = obj;
	}
    }


    protected synchronized void releaseObject(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
	Assert.that((obj.state & (Persistent.TRANSREAD|Persistent.TRANSWRITE))
		    != 0);
	CacheManager.unlink(obj); // remove object from transaction list

	if ((obj.state & (Persistent.SLOCKED|Persistent.XLOCKED)) != 0) {
	    obj.storage.unlock(obj, Protocol.lck_none);
	}	
	if ((obj.state & Persistent.NOTIFY) == 0) { 
	    mng.insertInCache(obj);
	}
	obj.state &= ~(Persistent.TRANSREAD | Persistent.TRANSWRITE |
		       Persistent.XLOCKED | Persistent.SLOCKED | 
		       Persistent.DIRTY | Persistent.NEW);
    }
	
    //
    // Destoroy all object references to make it possible to provide
    // more unaccessible objects to GC
    //
    protected void destroyObject(Persistent obj) {
	if ((obj.state & Persistent.DESTRUCTED) == 0) { 
	    try { 
		FieldDescriptor fd;
		for (fd = obj.desc.refFields; fd != null; fd = fd.nextRef) {
		    fd.field.set(obj, null);
		}
	    } catch(Exception x) { 
		throw new Error(x.getMessage());
	    }
	    obj.state |= Persistent.DESTRUCTED|Persistent.RAW;
	    obj.storage.throwObject(obj);
	}
    }

    //
    // This method is asynchronously called by server storag agent when
    // deterioration notification from server is received. 
    //
    protected synchronized void invalidateObject(Persistent obj) { 
	CacheManager mng = CacheManager.getCacheManager();
	if ((obj.state & Persistent.DESTRUCTED) == 0) {
	    if (obj.accessCount != 0 ||
		(obj.state & (Persistent.TRANSREAD|Persistent.TRANSWRITE))!=0)
	    { 
		//
		// It is not possible to detroy object immediatly, 
	        // just mark it as invalid and delay destruction till end
		// of access to the object or end of transaction.
		//
		obj.invalidated = true;
	    } else { 
		obj.invalidated = false;
		synchronized (mng) { 
		    mng.removeFromCache(obj);
		    destroyObject(obj);
		}
	    }
	}
	if ((obj.state & Persistent.NOTIFY) != 0) { 
	    CondEvent event = 
		(CondEvent)obj.storage.database.notifications.get(obj);
	    if (event != null) { 
		event.signal();
	    }
	}
    }

    //
    // Enable or disable notification on receving invalidation messages from
    // the server
    //
    protected synchronized void notifyOnModification(Persistent obj, 
						     boolean enabled)
    {
	CacheManager mng = CacheManager.getCacheManager();
	synchronized(mng) { 
	    if (enabled) { 
		obj.state |= Persistent.NOTIFY;
		if ((obj.state & Persistent.CACHED) == 0) { 
		    mng.removeFromCache(obj);
		} 
	    } else { 
		obj.state &= ~Persistent.NOTIFY;
		if (obj.accessCount == 0 
		    && (obj.state 
			& (Persistent.TRANSREAD|Persistent.TRANSWRITE)) == 0)
	        {
		    mng.insertInCache(obj);
		}
	    }
	}
    }



    //
    // By incrementing nested transaction count we delay implicit transaction 
    // commit until the moment when counter is explicitly decremented by
    // endNestedTransactionMethod()
    //
    public static void beginNestedTransaction() {
	CacheManager mng = (CacheManager)CacheManager.getCacheManager();
	synchronized (mng.globalCS) { 
	    mng.nestedTransactionCount += 1;
	}
    }

    public static void endNestedTransaction() { 
	CacheManager mng = (CacheManager)CacheManager.getCacheManager();
	synchronized (mng.globalCS) { 
	    if (--mng.nestedTransactionCount == 0) { 
		commitTransaction();
	    }
	}
    }
}



