#ifndef __CONFGRTR_H__
#define __CONFGRTR_H__

#include "goodsdlx.h"
#include "server.h"

#define GOODSRV_CFG_FILE_NAME  "goodsrv.cfg"

#define Kb 1024

GOODS_DLL_EXPORT extern unsigned init_map_file_size;
GOODS_DLL_EXPORT extern unsigned init_index_file_size;
// amount of memory after which allocation GC is initiated
GOODS_DLL_EXPORT extern unsigned gc_init_timeout;
GOODS_DLL_EXPORT extern unsigned gc_init_allocated;
GOODS_DLL_EXPORT extern unsigned gc_init_idle_period;
GOODS_DLL_EXPORT extern unsigned gc_init_min_allocated;
GOODS_DLL_EXPORT extern unsigned gc_response_timeout;
GOODS_DLL_EXPORT extern unsigned gc_grey_set_threshold;
GOODS_DLL_EXPORT extern unsigned max_data_file_size;
GOODS_DLL_EXPORT extern unsigned max_objects;

GOODS_DLL_EXPORT extern unsigned sync_log_writes;
GOODS_DLL_EXPORT extern unsigned permanent_backup;
GOODS_DLL_EXPORT extern unsigned trans_preallocated_log_size;
GOODS_DLL_EXPORT extern unsigned trans_max_log_size;
GOODS_DLL_EXPORT extern unsigned trans_wait_timeout;
GOODS_DLL_EXPORT extern unsigned trans_retry_timeout;
GOODS_DLL_EXPORT extern unsigned checkpoint_period;
GOODS_DLL_EXPORT extern unsigned dynamic_reclustering_limit;

GOODS_DLL_EXPORT extern unsigned lock_timeout;

GOODS_DLL_EXPORT extern unsigned page_pool_size;
 
GOODS_DLL_EXPORT extern char*    admin_telnet_port;
GOODS_DLL_EXPORT extern unsigned cluster_size;

GOODS_DLL_EXPORT extern char* index_file_name;
GOODS_DLL_EXPORT extern char* map_file_name;
GOODS_DLL_EXPORT extern char* data_file_name;
GOODS_DLL_EXPORT extern char* trans_file_name;
GOODS_DLL_EXPORT extern char* history_file_name;

struct param_binding {
     char const* param_name;
     unsigned*   param_ivalue;
     char const* param_meaning;	
     boolean     (*set_param)(dbs_server* srv, unsigned val);
     char**      param_svalue;
};

GOODS_DLL_EXPORT extern param_binding goodsrv_params[];

GOODS_DLL_EXPORT boolean set_max_data_file_size(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_max_objects(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_gc_init_timeout(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_gc_grey_set_threshold(dbs_server* srv,
					   	   unsigned val);
GOODS_DLL_EXPORT boolean set_gc_response_timeout(dbs_server* srv,
						 unsigned val);
GOODS_DLL_EXPORT boolean set_gc_init_allocated(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_gc_init_idle_period(dbs_server* srv,
						 unsigned val);
GOODS_DLL_EXPORT boolean set_gc_init_min_allocated(dbs_server* srv,
						   unsigned val);
GOODS_DLL_EXPORT boolean set_permanent_backup(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_checkpoint_period(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_max_log_size(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_preallocated_log_size(dbs_server* srv,
						   unsigned val);
GOODS_DLL_EXPORT boolean set_wait_timeout(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_retry_timeout(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_dynamic_reclustering_limit(dbs_server* srv,
							unsigned val);
GOODS_DLL_EXPORT boolean set_lock_timeout(dbs_server* srv, unsigned val);
GOODS_DLL_EXPORT boolean set_cluster_size(dbs_server* srv, unsigned val);

union option_value { 
    unsigned ivalue;
    char*    svalue;
};

GOODS_DLL_EXPORT boolean parse_option(char* buf, param_binding* &param,
				      option_value& u);
GOODS_DLL_EXPORT void read_goodsrv_configuration(const char* cfg_file_name);

#define SHOW_ALL "servers clients memory transaction classes"

GOODS_DLL_EXPORT void on_backup_completion(dbs_server&, file& backup_file,
					   boolean result);

struct trace_option { 
    const char* option;
    int         mask;
    const char* meaning;
};

GOODS_DLL_EXPORT extern trace_option trace_options_table[];

const int max_cmd_len = 256;

GOODS_DLL_EXPORT extern char      log_file_name[max_cmd_len];

GOODS_DLL_EXPORT extern char      monitor_options[max_cmd_len];
GOODS_DLL_EXPORT extern time_t    monitor_period;
GOODS_DLL_EXPORT extern semaphore monitor_sem;
GOODS_DLL_EXPORT extern event     monitor_term_event;

GOODS_DLL_EXPORT void task_proc monitor(void* arg);

GOODS_DLL_EXPORT boolean arg2int(char* arg, int& val);

GOODS_DLL_EXPORT void administrator_dialogue(char* database_config_name,
					     dbs_server& server);

GOODS_DLL_EXPORT void acceptAdminConnections(char* database_config_name,
                                             dbs_server& server);

GOODS_DLL_EXPORT int goodsrv(int argc, char **argv);

#endif
